"""Octomachinery CLI runner for GitHub Action environments."""

import asyncio
import logging

# pylint: disable=relative-beyond-top-level
from ...github.entities.action import GitHubAction
# pylint: disable=relative-beyond-top-level
from ...github.errors import GitHubActionError
# pylint: disable=relative-beyond-top-level
from ...github.models.action_outcomes import (
    ActionSuccess, ActionNeutral, ActionFailure,
)
# pylint: disable=relative-beyond-top-level
from ..config import BotAppConfig
# pylint: disable=relative-beyond-top-level
from ..routing.webhooks_dispatcher import (
    route_github_action_event,
)
# pylint: disable=relative-beyond-top-level
from ..runtime.context import RUNTIME_CONTEXT


logger = logging.getLogger(__name__)


async def process_github_action(config):
    """Schedule GitHub Action event for processing."""
    logger.info('Processing GitHub Action event...')

    github_action = GitHubAction(config.action)
    logger.info('GitHub Action=%r', config.action)

    await route_github_action_event(github_action)
    return ActionSuccess('GitHub Action has been processed')


def run():
    """Start up a server using CLI args for host and port."""
    config = BotAppConfig.from_dotenv()
    RUNTIME_CONTEXT.config = config  # pylint: disable=assigning-non-slot
    logging.basicConfig(
        level=logging.DEBUG
        if config.runtime.debug  # pylint: disable=no-member
        else logging.INFO,
    )
    try:
        processing_outcome = asyncio.run(process_github_action(config))
    except GitHubActionError as action_error:
        action_error.terminate_action()
    except KeyboardInterrupt:
        ActionNeutral('Action processing interrupted by user').raise_it()
    except Exception:  # pylint: disable=broad-except
        err_msg = 'Action processing failed unexpectedly'
        logger.exception(err_msg)
        ActionFailure(err_msg).raise_it()
    else:
        processing_outcome.raise_it()



"""Config schema for a GitHub App instance details."""
import environ


# pylint: disable=relative-beyond-top-level
from ...app.runtime.context import RUNTIME_CONTEXT
# pylint: disable=relative-beyond-top-level
from ..models.utils import SecretStr


@environ.config  # pylint: disable=too-few-public-methods
class GitHubAppIntegrationConfig:
    """GitHub App auth related config."""

    app_id = environ.var(name='GITHUB_APP_IDENTIFIER')
    private_key = environ.var(
        name='GITHUB_PRIVATE_KEY',
        converter=SecretStr,
    )
    webhook_secret = environ.var(
        None, name='GITHUB_WEBHOOK_SECRET',
        converter=lambda s: SecretStr(s) if s is not None else s,
    )

    @property
    def user_agent(self):  # noqa: D401
        """The User-Agent value to use when hitting GitHub API."""
        name = RUNTIME_CONTEXT.config.runtime.app_name
        version = RUNTIME_CONTEXT.config.runtime.app_version
        url = RUNTIME_CONTEXT.config.runtime.app_url
        return f'{name}/{version} (+{url})'



"""Web-server constructors."""

import asyncio
import logging

from aiohttp import web

# pylint: disable=relative-beyond-top-level
from ...github.api.app_client import GitHubApp
# pylint: disable=relative-beyond-top-level
from ..routing.webhooks_dispatcher import (
    route_github_webhook_event,
)
# pylint: disable=relative-beyond-top-level
from ..runtime.context import RUNTIME_CONTEXT


logger = logging.getLogger(__name__)


async def start_tcp_site(server_config, aiohttp_server_runner):
    """Return initialized and listening TCP site."""
    host, port = server_config.host, server_config.port
    aiohttp_tcp_site = web.TCPSite(aiohttp_server_runner, host, port)
    await aiohttp_tcp_site.start()
    logger.info(
        f' Serving on http://%s:%s/ '.center(50, '='),
        host, port,
    )
    return aiohttp_tcp_site


async def get_server_runner(http_handler):
    """Initialize server runner."""
    aiohttp_server = web.Server(http_handler)
    aiohttp_server_runner = web.ServerRunner(aiohttp_server)
    await aiohttp_server_runner.setup()
    return aiohttp_server_runner


async def run_forever(config):
    """Spawn an HTTP server in asyncio context."""
    async with GitHubApp(config.github) as github_app:
        logger.info('Starting the following GitHub App:')
        logger.info(
            '* app id: %s',
            github_app._config.app_id,  # pylint: disable=protected-access
        )
        logger.info(
            '* user agent: %s',
            github_app._config.user_agent,  # pylint: disable=protected-access
        )
        RUNTIME_CONTEXT.github_app = (  # pylint: disable=assigning-non-slot
            github_app
        )
        aiohttp_server_runner = await get_server_runner(
            route_github_webhook_event,
        )
        aiohttp_tcp_site = await start_tcp_site(
            config.server, aiohttp_server_runner,
        )

        if RUNTIME_CONTEXT.config.runtime.debug:
            logger.debug(
                'Running a GitHub App under env=%s',
                RUNTIME_CONTEXT.config.runtime.env,
            )

        try:
            await asyncio.get_event_loop().create_future()  # block
        except asyncio.CancelledError:
            logger.info(' Stopping the server '.center(50, '='))
            await aiohttp_tcp_site.stop()



"""Octomachinery CLI runner."""

import asyncio
import logging
import sys
from typing import Optional

import attr

# pylint: disable=relative-beyond-top-level
from ..config import BotAppConfig
# pylint: disable=relative-beyond-top-level
from ..runtime.context import RUNTIME_CONTEXT
# pylint: disable=relative-beyond-top-level
from .config import WebServerConfig
# pylint: disable=relative-beyond-top-level
from .machinery import run_forever as run_server_forever


logger = logging.getLogger(__name__)


def run(
        *,
        name: Optional[str] = None,
        version: Optional[str] = None,
        url: Optional[str] = None,
):
    """Start up a server using CLI args for host and port."""
    config = BotAppConfig.from_dotenv(
        app_name=name,
        app_version=version,
        app_url=url,
    )
    if len(sys.argv) > 2:
        config = attr.evolve(
            config,
            server=WebServerConfig(*sys.argv[1:3]),
        )
    RUNTIME_CONTEXT.config = config  # pylint: disable=assigning-non-slot
    if config.runtime.debug:  # pylint: disable=no-member
        logging.basicConfig(level=logging.DEBUG)

        logger.debug(
            ' App version: %s '.center(50, '='),
            config.runtime.app_version,
        )
    else:
        logging.basicConfig(level=logging.INFO)

    try:
        asyncio.run(run_server_forever(config))
    except KeyboardInterrupt:
        logger.info(' Exiting the app '.center(50, '='))



"""GitHub webhook events dispatching logic."""

import asyncio
from functools import wraps
from http import HTTPStatus
import logging

from aiohttp import web
from gidgethub import BadRequest, ValidationFailure

# pylint: disable=relative-beyond-top-level,import-error
from ..runtime.context import RUNTIME_CONTEXT
from . import dispatch_event


__all__ = ('route_github_action_event', 'route_github_webhook_event', )


logger = logging.getLogger(__name__)


EVENT_LOG_TMPL = (
    'Got a{}valid X-GitHub-Event=%s '
    'with X-GitHub-Delivery=%s '
    'and X-Hub-Signature=%s'
)

EVENT_INVALID_CHUNK = 'n in'
EVENT_VALID_CHUNK = ' '

EVENT_LOG_VALID_MSG = EVENT_LOG_TMPL.format(EVENT_VALID_CHUNK)
EVENT_LOG_INVALID_MSG = EVENT_LOG_TMPL.format(EVENT_INVALID_CHUNK)


async def get_event_from_request(request):
    """Retrieve Event out of HTTP request if it's valid."""
    webhook_event_signature = request.headers.get(
        'X-Hub-Signature', '<MISSING>',
    )
    try:
        event = await RUNTIME_CONTEXT.github_app.event_from_request(request)
    except ValidationFailure as no_signature_exc:
        logger.info(
            EVENT_LOG_INVALID_MSG,
            request.headers.get('X-GitHub-Event'),
            request.headers.get('X-GitHub-Delivery'),
            webhook_event_signature,
        )
        raise web.HTTPForbidden from no_signature_exc
    else:
        logger.info(
            EVENT_LOG_VALID_MSG,
            event.event,
            event.delivery_id,
            webhook_event_signature,
        )
        return event


def validate_allowed_http_methods(*allowed_methods):
    """Block disallowed HTTP methods."""
    if not allowed_methods:
        allowed_methods = {'POST'}
    else:
        allowed_methods = set(allowed_methods)

    def decorator(wrapped_function):
        @wraps(wrapped_function)
        async def wrapper(request):
            if request.method not in allowed_methods:
                raise web.HTTPMethodNotAllowed(
                    method=request.method,
                    allowed_methods=allowed_methods,
                ) from BadRequest(HTTPStatus.METHOD_NOT_ALLOWED)
            return await wrapped_function(request)
        return wrapper
    return decorator


@validate_allowed_http_methods('POST')
async def route_github_webhook_event(request):
    """Dispatch incoming webhook events to corresponsing handlers."""
    github_app = RUNTIME_CONTEXT.github_app

    event = await get_event_from_request(request)

    await asyncio.sleep(1)  # Give GitHub a sec to deal w/ eventual consistency
    async with github_app.github_app_client:
        github_installation = await github_app.get_installation(event)
        github_installation_client = (
            github_installation.github_installation_client
        )
        async with github_installation_client as gh_install_client:
            # pylint: disable=assigning-non-slot
            RUNTIME_CONTEXT.app_installation_client = gh_install_client
            await dispatch_event(event)
    return web.Response(
        text=f'OK: GitHub event received. It is {event.event!s} ({event!r})',
    )


async def route_github_action_event(github_action):
    """Dispatch a GitHub action event to corresponsing handlers."""
    async with github_action.github_installation_client as gh_install_client:
        # pylint: disable=assigning-non-slot
        RUNTIME_CONTEXT.app_installation_client = gh_install_client
        await dispatch_event(github_action.event)



"""GitHub App Installation wrapper."""

import logging

import attr

# pylint: disable=relative-beyond-top-level
from ...app.runtime.context import RUNTIME_CONTEXT
# pylint: disable=relative-beyond-top-level
from ..api.client import GitHubAPIClient
# pylint: disable=relative-beyond-top-level
from ..api.tokens import GitHubOAuthToken
# pylint: disable=relative-beyond-top-level
from ..models import (
    GitHubAppInstallation as GitHubAppInstallationModel,
    GitHubInstallationAccessToken,
)


logger = logging.getLogger(__name__)


@attr.dataclass
class GitHubAppInstallation:
    """GitHub App Installation API wrapper."""

    _metadata: GitHubAppInstallationModel
    """A GitHub Installation metadata from GitHub webhook."""
    _token: GitHubInstallationAccessToken = attr.ib(init=False, default=None)
    """A GitHub Installation token for GitHub API."""

    @property
    def token(self):
        """Return GitHub App Installation access token."""
        try:
            if self._token.expired:
                return None

            return GitHubOAuthToken(str(self._token))
        except TypeError:
            return None

    async def retrieve_access_token(self):
        """Retrieve installation access token from GitHub API."""
        async with RUNTIME_CONTEXT.github_app.github_app_client as gh_api:
            self._token = GitHubInstallationAccessToken(**(await gh_api.post(
                self._metadata.access_tokens_url,
                data=b'',
                accept='application/vnd.github.machine-man-preview+json',
            )))
        return self._token

    @property
    def github_installation_client(self):  # noqa: D401
        """The GitHub App client with an async CM interface."""
        try:
            token = GitHubOAuthToken(self._token.token)
        except AttributeError:
            token = None
        return GitHubAPIClient(token)



"""GitHub App API client."""

from collections import defaultdict
from contextlib import AbstractAsyncContextManager
import logging
import time
import types
import typing

from aiohttp.client_exceptions import ClientConnectorError
import attr
from gidgethub.sansio import Event
import jwt

# pylint: disable=relative-beyond-top-level
from ...app.runtime.context import RUNTIME_CONTEXT
# pylint: disable=relative-beyond-top-level
from ...utils.asynctools import (
    amap, dict_to_kwargs_cb,
)
# pylint: disable=relative-beyond-top-level
from ..config.app import GitHubAppIntegrationConfig
# pylint: disable=relative-beyond-top-level
from ..entities.app_installation import GitHubAppInstallation
# pylint: disable=relative-beyond-top-level
from ..models import GitHubAppInstallation as GitHubAppInstallationModel
from .client import GitHubAPIClient
from .tokens import GitHubJWTToken


logger = logging.getLogger(__name__)


GH_INSTALL_EVENTS = {'integration_installation', 'installation'}


@attr.dataclass
class GitHubApp(AbstractAsyncContextManager):
    """GitHub API wrapper."""

    _config: GitHubAppIntegrationConfig

    def __attrs_post_init__(self):
        """Initialize installations store."""
        # pylint: disable=attribute-defined-outside-init
        self._installations = defaultdict(dict)

    async def event_from_request(self, request):
        """Get an event object out of HTTP request."""
        event = Event.from_http(
            request.headers,
            await request.read(),
            secret=self._config.webhook_secret,
        )
        await self.pre_process_webhook_event(event)
        return event

    async def pre_process_webhook_event(self, event):
        """Get an event object out of HTTP request."""
        action = event.data.get('action')
        if event.event in GH_INSTALL_EVENTS and action == 'created':
            await self.add_installation(event)

    async def __aenter__(self) -> 'GitHubApp':
        """Store all installations data before starting."""
        RUNTIME_CONTEXT.github_app = self
        # pylint: disable=attribute-defined-outside-init
        try:
            self._installations = await self.get_installations()
        except ClientConnectorError as client_error:
            logger.info('It looks like the GitHub API is offline...')
            logger.error(
                'The following error has happened while trying to grab '
                'installations list: %s',
                client_error,
            )
            self._installations = defaultdict(dict)

        logger.info('This GitHub App is installed into:')
        # pylint: disable=protected-access
        for install_id, install_val in self._installations.items():
            logger.info(
                '* Installation id %s (expires at %s, installed to %s)',
                install_id,
                install_val._token.expires_at,
                install_val._metadata.account['login'],
            )

        return self

    async def __aexit__(
            self,
            exc_type: typing.Optional[typing.Type[BaseException]],
            exc_val: typing.Optional[BaseException],
            exc_tb: typing.Optional[types.TracebackType],
    ) -> typing.Optional[bool]:
        """Wipe out the installation store."""
        # pylint: disable=attribute-defined-outside-init
        self._installations = defaultdict(dict)

    @property
    def gh_jwt(self):
        """Generate app's JSON Web Token, valid for 60 seconds."""
        now = int(time.time())
        payload = {
            'iat': now,
            'exp': now + 60,
            'iss': self._config.app_id,
        }
        token = jwt.encode(
            payload,
            key=self._config.private_key,
            algorithm='RS256',
        ).decode('utf-8')
        return GitHubJWTToken(token)

    @property
    def github_app_client(self):  # noqa: D401
        """The GitHub App client with an async CM interface."""
        return GitHubAPIClient(self.gh_jwt)

    async def add_installation(self, event):
        """Retrieve an installation creds from store."""
        install = event.data['installation']
        install_id = install['id']
        self._installations[install_id] = GitHubAppInstallation(install)
        await self._installations[install_id].retrieve_access_token()
        return self._installations[install_id]

    async def get_installation(self, event):
        """Retrieve an installation creds from store."""
        if event.event == 'ping':
            return GitHubAppInstallation(None)

        install_id = event.data['installation']['id']
        app_installation = self._installations.get(install_id)

        # pylint: disable=assigning-non-slot
        RUNTIME_CONTEXT.app_installation = (
            app_installation
        )

        return app_installation

    async def get_installations(self):
        """Retrieve all installations with access tokens via API."""
        installations = defaultdict(dict)
        async with self.github_app_client as gh_api:
            async for install in amap(
                    dict_to_kwargs_cb(GitHubAppInstallationModel),
                    gh_api.getiter(
                        '/app/installations',
                        preview_api_version='machine-man',
                    ),
            ):
                installations[install.id] = GitHubAppInstallation(install)
                await installations[install.id].retrieve_access_token()
        return installations



"""A low-level GitHub API client."""

from contextlib import AbstractAsyncContextManager
import types
import typing

import aiohttp
import attr

# pylint: disable=relative-beyond-top-level
from ...app.runtime.context import RUNTIME_CONTEXT
from .raw_client import RawGitHubAPI
from .tokens import GitHubToken
from .utils import mark_uninitialized_in_repr


@mark_uninitialized_in_repr
@attr.dataclass
class GitHubAPIClient(AbstractAsyncContextManager):
    """A client to the GitHub API with an asynchronous CM support."""

    _github_token: GitHubToken
    _external_session: typing.Optional[aiohttp.ClientSession] = (
        attr.ib(default=None)
    )
    """A session created externally."""
    _current_session: aiohttp.ClientSession = attr.ib(init=False, default=None)
    """A session created per CM if there's no external one."""
    _api_client: RawGitHubAPI = attr.ib(init=False, default=None)
    """A Gidgethub client for GitHub API."""

    def __attrs_post_init__(self):
        """Gidgethub API client instance initializer."""
        try:
            self._api_client = RawGitHubAPI(
                token=self._github_token,
                session=self._open_session(),
                user_agent=RUNTIME_CONTEXT.config.github.user_agent,
            )
        except (AttributeError, TypeError):
            pass

    @property
    def is_initialized(self):
        """Return GitHub token presence."""
        return self._github_token is not None

    def _open_session(self) -> aiohttp.ClientSession:
        """Return a session to use with GitHub API."""
        assert self._current_session is None
        self._current_session = (
            aiohttp.ClientSession() if self._external_session is None
            else self._external_session
        )
        return self._current_session

    async def _close_session(self) -> None:
        """Free up the current session."""
        assert self._current_session is not None
        if self._external_session is None:
            await self._current_session.close()
        self._current_session = None

    async def __aenter__(self) -> RawGitHubAPI:
        """Return a GitHub API wrapper."""
        return self._api_client

    async def __aexit__(
            self,
            exc_type: typing.Optional[typing.Type[BaseException]],
            exc_val: typing.Optional[BaseException],
            exc_tb: typing.Optional[types.TracebackType],
    ) -> typing.Optional[bool]:
        """Close the current session resource."""
        await self._close_session()
